// Copyright 2011 Julian Phillips.  All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package py

// #include "utils.h"
//import "C"

import "unsafe"

func packValues(values []interface{}) ([]unsafe.Pointer, error) {
	cValues := make([]unsafe.Pointer, len(values))
	for i, value := range values {
		switch v := value.(type) {
		case *string:
			cValues[i] = unsafe.Pointer(new(*int8))
		case *Object:
			cValues[i] = unsafe.Pointer(new(*_PyObject))
		case *int:
			cValues[i] = unsafe.Pointer(new(int32))
		case *int8:
			cValues[i] = unsafe.Pointer(new(int8))
		case *int16:
			cValues[i] = unsafe.Pointer(new(int16))
		case *int32:
			cValues[i] = unsafe.Pointer(new(int32))
		case *int64:
			cValues[i] = unsafe.Pointer(new(int64))
		case *uint:
			cValues[i] = unsafe.Pointer(new(uint32))
		case *uint8:
			cValues[i] = unsafe.Pointer(new(uint8))
		case *uint16:
			cValues[i] = unsafe.Pointer(new(uint16))
		case *uint32:
			cValues[i] = unsafe.Pointer(new(uint32))
		case *uint64:
			cValues[i] = unsafe.Pointer(new(uint64))
		case *float32:
			cValues[i] = unsafe.Pointer(new(float32))
		case *float64:
			cValues[i] = unsafe.Pointer(new(float64))
		default:
			return nil, TypeError.Err("Unsupported type: %T", v)
		}
	}
	return cValues, nil
}

func unpackValues(cValues []unsafe.Pointer, values []interface{}) error {
	for i, value := range values {
		switch v := value.(type) {
		case *string:
			*v = _GoString(*(**int8)(cValues[i]))
		case *Object:
			*v = newObject(*(**_PyObject)(cValues[i]))
		case *int:
			*v = int(*(*int32)(cValues[i]))
		case *int8:
			*v = int8(*(*int8)(cValues[i]))
		case *int16:
			*v = int16(*(*int16)(cValues[i]))
		case *int32:
			*v = int32(*(*int32)(cValues[i]))
		case *int64:
			*v = int64(*(*int64)(cValues[i]))
		case *uint:
			*v = uint(*(*uint32)(cValues[i]))
		case *uint8:
			*v = uint8(*(*uint8)(cValues[i]))
		case *uint16:
			*v = uint16(*(*uint16)(cValues[i]))
		case *uint32:
			*v = uint32(*(*uint32)(cValues[i]))
		case *uint64:
			*v = uint64(*(*uint64)(cValues[i]))
		case *float32:
			*v = float32(*(*float32)(cValues[i]))
		case *float64:
			*v = float64(*(*float64)(cValues[i]))
		default:
			return TypeError.Err("Unsupported type: %T", v)
		}
	}
	return nil
}

func ParseTuple(args *Tuple, format string, values ...interface{}) error {
	if args == nil {
		return AssertionError.Err("ParseTuple: args was nil")
	}

	cv := (**byte)(nil)
	cValues, err := packValues(values)
	if err != nil {
		return nil
	}

	if len(cValues) > 0 {
		cv = (**byte)(unsafe.Pointer(&cValues[0]))
	}

	f := _CString(format)
	defer _free(unsafe.Pointer(f))

	ret := _doParseTuple(c(args), f, cv, int32(len(cValues)))
	if ret == 0 {
		return exception()
	}

	return unpackValues(cValues, values)
}

func ParseTupleAndKeywords(args *Tuple, kw *Dict, format string, kwlist []string, values ...interface{}) error {
	if args == nil {
		return AssertionError.Err("ParseTupleAndKeywords: args was nil")
	}

	cv := (**byte)(nil)
	cValues, err := packValues(values)
	if err != nil {
		return nil
	}

	if len(cValues) > 0 {
		cv = (**byte)(unsafe.Pointer(&cValues[0]))
	}

	f := _CString(format)
	defer _free(unsafe.Pointer(f))

	klist := make([]*int8, len(kwlist)+1)

	for i, k := range kwlist {
		klist[i] = _CString(k)
		defer _free(unsafe.Pointer(klist[i]))
	}

	ret := _doParseTupleKwds(c(args), c(kw), f, &klist[0], cv, int32(len(cValues)))
	if ret == 0 {
		return exception()
	}

	return unpackValues(cValues, values)
}

func BuildValue(format string, values ...interface{}) (Object, error) {
	cValues := make([]_ArgValue, len(values))
	for i, value := range values {
		switch v := value.(type) {
		case string:
			s := _CString(v)
			defer _free(unsafe.Pointer(s))
			cValues[i]._type = &*_ffi_type_pointer
			cValues[i].value = (*byte)(unsafe.Pointer(&s))
		case Object:
			b := v.Base()
			cValues[i]._type = &*_ffi_type_pointer
			cValues[i].value = (*byte)(unsafe.Pointer(&b))
		case int:
			iv := int32(v)
			cValues[i]._type = &*_ffi_type_sint
			cValues[i].value = (*byte)(unsafe.Pointer(&iv))
		case int8:
			iv := int8(v)
			cValues[i]._type = &*_ffi_type_sint8
			cValues[i].value = (*byte)(unsafe.Pointer(&iv))
		case int16:
			iv := int16(v)
			cValues[i]._type = &*_ffi_type_sint16
			cValues[i].value = (*byte)(unsafe.Pointer(&iv))
		case int32:
			iv := int32(v)
			cValues[i]._type = &*_ffi_type_sint32
			cValues[i].value = (*byte)(unsafe.Pointer(&iv))
		case int64:
			iv := int64(v)
			cValues[i]._type = &*_ffi_type_sint64
			cValues[i].value = (*byte)(unsafe.Pointer(&iv))
		case uint:
			iv := uint32(v)
			cValues[i]._type = &*_ffi_type_uint
			cValues[i].value = (*byte)(unsafe.Pointer(&iv))
		case uint8:
			iv := uint8(v)
			cValues[i]._type = &*_ffi_type_uint8
			cValues[i].value = (*byte)(unsafe.Pointer(&iv))
		case uint16:
			iv := uint16(v)
			cValues[i]._type = &*_ffi_type_uint16
			cValues[i].value = (*byte)(unsafe.Pointer(&iv))
		case uint32:
			iv := uint32(v)
			cValues[i]._type = &*_ffi_type_uint32
			cValues[i].value = (*byte)(unsafe.Pointer(&iv))
		case uint64:
			iv := uint64(v)
			cValues[i]._type = &*_ffi_type_uint64
			cValues[i].value = (*byte)(unsafe.Pointer(&iv))
		case float32:
			fv := float32(v)
			cValues[i]._type = &*_ffi_type_float
			cValues[i].value = (*byte)(unsafe.Pointer(&fv))
		case float64:
			fv := float64(v)
			cValues[i]._type = &*_ffi_type_double
			cValues[i].value = (*byte)(unsafe.Pointer(&fv))
		default:
			return nil, TypeError.Err("Unsupported type: %T", v)
		}
	}
	f := _CString(format)
	defer _free(unsafe.Pointer(f))
	ret := _doBuildValue(f, &cValues[0], int32(len(cValues)))
	if ret == nil {
		return nil, exception()
	}
	return newObject(ret), nil
}
